from pathlib import Path

import pandas as pd
import requests
import torch
from eyepacs import EyePACS

from EyeQ_process import process


class EyeQ(EyePACS):

    # Create EyeQ dataset class attributes
    image_extension = ".jpg"
    result_label = "quality"
    classes_names = ["Gradable", "Usable", "Ungradable"]

    def __init__(
        self,
        targ_dir: Path = "",  # also called root
        split: str = "train",
        transform=None,
        target_transform=None,
    ) -> None:
        super().__init__(targ_dir, split, transform, target_transform)

        self.__preprocess()

        self.label_csv_name = f"Label_EyeQ_{split}.csv"

        self.__download_results()

        self.df_labels = pd.read_csv(self.data_path / self.label_csv_name, sep=",")

        self.df_labels["image"] = self.df_labels["image"].apply(
            lambda S: S.strip(".jpeg")
        )

    def __download_results(
        self,
    ) -> None:
        # Download EyeQ results
        with open(self.data_path / self.label_csv_name, "wb") as f:
            request = requests.get(
                f"https://raw.githubusercontent.com/HzFu/EyeQ/master/data/{self.label_csv_name}"
            )
            print(f"Downloading Label_EyeQ_{self.split}.csv...")
            f.write(request.content)

    def __preprocess(
        self,
    ) -> None:
        image_list = [str(img) for img in self.paths]
        save_path = str(self.data_path / self.split)

        process(image_list, save_path, self.image_extension)

        self.paths = list(Path(save_path).glob("*" + self.image_extension))


if __name__ == "__main__":

    from torchvision import transforms

    data_transform = transforms.Compose(
        [
            # Resize and Crop the images to 640x640
            transforms.Resize(EyeQ.image_size),
            transforms.CenterCrop(EyeQ.image_size),
            # Turn the image into a torch.Tensor
            transforms.ToTensor(),  # this also converts all pixel values from 0 to 255 to be between 0.0 and 1.0
            transforms.Normalize(EyeQ.mean, EyeQ.std),
        ]
    )

    # Setup path to data folder
    data_path = Path("~/.torchvision/")
    image_path = data_path / "eyeq"

    split = "train"

    dataset = EyeQ(image_path, split=split, transform=data_transform)  # train_data
    n = len(dataset)
    print(f"EyeQ, split {split}, has {n} samples.")

    ################################################################################################

    import matplotlib.pyplot as plt
    import torchvision

    def display_image(
        dataset: torchvision.datasets.VisionDataset, targ_sample: int
    ) -> None:
        targ_image, targ_label = dataset[targ_sample][0], dataset[targ_sample][1]

        # Adjust image tensor shape for plotting (rearrange the order of dimensions)
        # [color_channels, height, width] -> [height, width, color_channels]
        targ_image_adjust = (
            targ_image.permute(1, 2, 0) if torch.is_tensor(targ_image) else targ_image
        )

        # Setup plot
        plt.imshow(targ_image_adjust)
        plt.axis("off")
        title = f"class: {EyeQ.classes_names[targ_label]}"
        if torch.is_tensor(targ_image_adjust):
            title = title + f"\nshape: {targ_image_adjust.shape}"
        plt.title(title)

        # Plot the adjusted sample
        plt.show()

    display_image(dataset, 2)  # 13_left

    ################################################################################################

    from torch.utils.data import DataLoader

    train_dataloader_custom = DataLoader(
        dataset=dataset,  # use custom created train Dataset
        batch_size=1,  # how many samples per batch?
        num_workers=0,  # how many subprocesses to use for data loading? (higher = more)
        shuffle=True,  # shuffle the data?
    )

    # Get image and label from custom DataLoader
    img_custom = None
    while img_custom is None:
        try:
            img_custom, label_custom = next(iter(train_dataloader_custom))
        except (StopIteration, TypeError):
            break
        except:
            pass

    # Batch size will now be 1, try changing the batch_size parameter above and see what happens
    print(
        f"Image shape: {img_custom.shape} -> [batch_size, color_channels, height, width]"
    )
    print(f"Label shape: {label_custom.shape}")



import glob
import os
import zipfile
from pathlib import Path
from typing import Dict, List, Tuple

import pandas as pd
import torch

from kaggle.api.kaggle_api_extended import KaggleApi
from PIL import Image
from torchvision.datasets import VisionDataset
from torchvision.datasets.utils import check_integrity, extract_archive


def check_exists(root: Path, resources: List[Tuple[str, str]]) -> bool:
    return all(check_integrity(root / file, md5) for file, md5 in resources)


class EyePACS(VisionDataset):

    image_size = 640
    mean = [0.485, 0.456, 0.406]
    std = [0.229, 0.224, 0.225]

    # Create EyePACS dataset class attributes
    image_extension = ".jpeg"
    result_label = "level"
    label_csv_name = "trainLabels.csv"
    classes_names = ["No DR", "Mild", "Moderate", "Severe", "Proliferative DR"]

    main_zip_name = "diabetic-retinopathy-detection.zip"
    main_zip_hash = "596cf4ecabf92e5e621ac7e9f9181471"

    def __init__(
        self,
        targ_dir: Path = "",  # also called root
        split: str = "train",
        transform=None,
        target_transform=None,
    ) -> None:
        assert split in ["train", "val", "test"]

        if isinstance(targ_dir, str):
            targ_dir = Path(targ_dir)

        targ_dir = targ_dir.expanduser()

        self.split = split

        # Setup train or testing path as root
        images_root = targ_dir / split
        # Setup transforms
        super().__init__(
            images_root, transform=transform, target_transform=target_transform
        )

        self.data_path = self.root.parent

        self.api = KaggleApi()
        self.api.authenticate()

        self.__download_dataset()
        self.__extract_dataset()

        # Get all image paths
        self.paths = list(Path(images_root).glob("*" + self.image_extension))

        self.df_labels = pd.read_csv(self.data_path / self.label_csv_name, sep=",")

    def __download_dataset(
        self,
    ) -> None:
        main_zip_resource = [(self.main_zip_name, self.main_zip_hash)]
        if check_exists(self.data_path, main_zip_resource):
            return

        # If the image folder doesn't exist, download it and prepare it...
        if not self.root.is_dir():
            self.root.mkdir(parents=True, exist_ok=True)

            # Download diabetic-retinopathy-detection.zip
            print("Downloading dataset (88.29gbs), this may take a while...")

            self.api.competition_download_files(
                "diabetic-retinopathy-detection", path=self.data_path
            )

            if not check_exists(self.data_path, main_zip_resource):
                raise OSError(
                    f"File {self.main_zip_name} has not been downloaded correctly."
                )

    def __extract_dataset(
        self,
    ) -> None:
        # Unzip diabetic-retinopathy-detection.zip
        # extract_archive(self.data_path / self.main_zip_name, self.data_path)

        zip_prefix = f"{self.split}.zip."

        with zipfile.ZipFile(self.data_path / self.main_zip_name, "r") as zip_ref:
            for file in zip_ref.namelist():
                if file.startswith(zip_prefix):
                    zip_ref.extract(file, self.data_path)
                if not file.startswith(("train.zip", "test.zip")):
                    zip_ref.extract(file, self.data_path)

        # N number of parts
        parts = glob.glob(str(self.data_path / (zip_prefix + "*")))
        n = len(parts)

        # Concatenate
        with open(self.data_path / f"{self.split}.zip", "wb") as outfile:
            for i in range(1, n + 1):
                filename = zip_prefix + str(i).zfill(3)
                with open(self.data_path / filename, "rb") as infile:
                    outfile.write(infile.read())

        for filename in os.listdir(self.data_path):
            if filename.startswith(zip_prefix):
                os.remove(self.data_path / filename)

        # Extract
        for file_path in os.listdir(self.data_path):
            if os.path.isfile(os.path.join(self.data_path, file_path)):
                main_zip_path = self.data_path / file_path
                if os.path.basename(main_zip_path) == self.main_zip_name:
                    continue
                if main_zip_path.suffix == ".zip":
                    print(f"Extracting {main_zip_path} to {self.data_path}...")
                    extract_archive(self.data_path / file_path, self.data_path)

        os.remove(self.data_path / f"{self.split}.zip")

    def load_image(self, index: int) -> Image.Image:
        "Opens an image via a path and returns it."
        image_name = self.df_labels["image"].iloc[index]
        image_name = image_name + self.image_extension
        image_path = self.root / image_name

        return Image.open(image_path)

    # Overwrites the __len__() method (optional but recommended for subclasses of torch.utils.data.Dataset)
    def __len__(self) -> int:
        "Returns the total number of samples."
        return len(self.paths)

    # Overwrites the __getitem__() method (required for subclasses of torch.utils.data.Dataset)
    def __getitem__(self, index: int) -> Tuple[torch.Tensor, int]:
        "Returns one sample of data, data and label (X, y)."

        img = self.load_image(index)
        class_idx = self.df_labels[self.result_label].iloc[index]

        # Transform if necessary
        if self.transform:
            img = self.transform(img)

        return img, class_idx  # return data, label (X, y)


if __name__ == "__main__":

    from torchvision import transforms

    data_transform = transforms.Compose(
        [
            # Resize and Crop the images to 640x640
            transforms.Resize(EyePACS.image_size),
            transforms.CenterCrop(EyePACS.image_size),
            # Turn the image into a torch.Tensor
            transforms.ToTensor(),  # this also converts all pixel values from 0 to 255 to be between 0.0 and 1.0
            transforms.Normalize(EyePACS.mean, EyePACS.std),
        ]
    )

    # Setup path to data folder
    data_path = Path("~/.torchvision/")
    image_path = data_path / "eyepacs"

    split = "train"

    dataset = EyePACS(image_path, split=split, transform=data_transform)  # train_data
    n = len(dataset)
    print(f"EyePACS, split {split}, has {n} samples.")

    ################################################################################################

    import matplotlib.pyplot as plt
    import torchvision

    def display_image(
        dataset: torchvision.datasets.VisionDataset, targ_sample: int
    ) -> None:
        targ_image, targ_label = dataset[targ_sample][0], dataset[targ_sample][1]

        # Adjust image tensor shape for plotting (rearrange the order of dimensions)
        # [color_channels, height, width] -> [height, width, color_channels]
        targ_image_adjust = (
            targ_image.permute(1, 2, 0) if torch.is_tensor(targ_image) else targ_image
        )

        # Setup plot
        plt.imshow(targ_image_adjust)
        plt.axis("off")
        title = f"class: {EyePACS.classes_names[targ_label]}"
        if torch.is_tensor(targ_image_adjust):
            title = title + f"\nshape: {targ_image_adjust.shape}"
        plt.title(title)

        # Plot the adjusted sample
        plt.show()

    display_image(dataset, 2)  # 13_left

    ################################################################################################

    from torch.utils.data import DataLoader

    train_dataloader_custom = DataLoader(
        dataset=dataset,  # use custom created train Dataset
        batch_size=1,  # how many samples per batch?
        num_workers=0,  # how many subprocesses to use for data loading? (higher = more)
        shuffle=True,  # shuffle the data?
    )

    # Get image and label from custom DataLoader
    img_custom = None
    while img_custom is None:
        try:
            img_custom, label_custom = next(iter(train_dataloader_custom))
        except (StopIteration, TypeError):
            break
        except:
            pass

    # Batch size will now be 1, try changing the batch_size parameter above and see what happens
    print(
        f"Image shape: {img_custom.shape} -> [batch_size, color_channels, height, width]"
    )
    print(f"Label shape: {label_custom.shape}")



import numpy as np
import os
import cv2


def imread(file_path, c=None):
    if c is None:
        im = cv2.imread(file_path)
    else:
        im = cv2.imread(file_path, c)

    if im is None:
        raise 'Can not read image'

    if im.ndim == 3 and im.shape[2] == 3:
        im = cv2.cvtColor(im, cv2.COLOR_BGR2RGB)
    return im


def imwrite(file_path, image):
    if image.ndim == 3 and image.shape[2] == 3:
        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
    cv2.imwrite(file_path, image)


def fold_dir(folder):
    if not os.path.exists(folder):
        os.makedirs(folder)
    return folder


def get_mask_BZ(img):
    if img.ndim==3:
        gray_img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    else:
        gray_img = img
    threhold = np.mean(gray_img)/3-5
    _, mask = cv2.threshold(gray_img, max(0,threhold), 1, cv2.THRESH_BINARY)
    nn_mask = np.zeros((mask.shape[0]+2,mask.shape[1]+2),np.uint8)
    new_mask = (1-mask).astype(np.uint8)
    #  cv::floodFill(Temp, Point(0, 0), Scalar(255));
    # _,new_mask,_,_ = cv2.floodFill(new_mask, nn_mask, [(0, 0),(0,new_mask.shape[0])], (0), cv2.FLOODFILL_MASK_ONLY)
    _,new_mask,_,_ = cv2.floodFill(new_mask, nn_mask, (0,0), (0), cv2.FLOODFILL_MASK_ONLY)
    _,new_mask,_,_ = cv2.floodFill(new_mask, nn_mask, (new_mask.shape[1]-1,new_mask.shape[0]-1), (0), cv2.FLOODFILL_MASK_ONLY)
    mask = mask + new_mask
    kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (20,  20))
    mask = cv2.erode(mask, kernel)
    mask = cv2.dilate(mask, kernel)
    return mask


def _get_center_by_edge(mask):
    center=[0,0]
    x=mask.sum(axis=1)
    center[0]=np.where(x>x.max()*0.95)[0].mean()
    x=mask.sum(axis=0)
    center[1]=np.where(x>x.max()*0.95)[0].mean()
    return center


def _get_radius_by_mask_center(mask,center):
    mask=mask.astype(np.uint8)
    ksize=max(mask.shape[1]//400*2+1,3)
    kernel=cv2.getStructuringElement(cv2.MORPH_ELLIPSE,(ksize,ksize))
    mask=cv2.morphologyEx(mask, cv2.MORPH_GRADIENT, kernel)
    # radius=
    index=np.where(mask>0)
    d_int=np.sqrt((index[0]-center[0])**2+(index[1]-center[1])**2)
    b_count=np.bincount(np.ceil(d_int).astype(np.int))
    radius=np.where(b_count>b_count.max()*0.995)[0].max()
    return radius


def _get_circle_by_center_bbox(shape,center,bbox,radius):
    center_mask=np.zeros(shape=shape).astype('uint8')
    tmp_mask=np.zeros(shape=bbox[2:4])
    center_tmp=(int(center[0]),int(center[1]))
    center_mask=cv2.circle(center_mask,center_tmp[::-1],int(radius),(1),-1)
    # center_mask[bbox[0]:bbox[0]+bbox[2],bbox[1]:bbox[1]+bbox[3]]=tmp_mask
    # center_mask[bbox[0]:min(bbox[0]+bbox[2],center_mask.shape[0]),bbox[1]:min(bbox[1]+bbox[3],center_mask.shape[1])]=tmp_mask
    return center_mask


def get_mask(img):
    if img.ndim ==3:
        #raise 'image dim is not 3'
        g_img=cv2.cvtColor(img,cv2.COLOR_RGB2GRAY)
    elif img.ndim == 2:
        g_img =img.copy()
    else:
        raise 'image dim is not 1 or 3'
    h,w = g_img.shape
    shape=g_img.shape[0:2]
    g_img = cv2.resize(g_img,(0,0),fx = 0.5,fy = 0.5)
    tg_img=cv2.normalize(g_img, None, 0, 255, cv2.NORM_MINMAX)
    tmp_mask=get_mask_BZ(tg_img)
    center=_get_center_by_edge(tmp_mask)
    #bbox=_get_bbox_by_mask(tmp_mask)
    radius=_get_radius_by_mask_center(tmp_mask,center)
    #resize back
    center = [center[0]*2,center[1]*2]
    radius = int(radius*2)
    s_h = max(0,int(center[0] - radius))
    s_w = max(0, int(center[1] - radius))
    bbox = (s_h, s_w, min(h-s_h,2 * radius), min(w-s_w,2 * radius))
    tmp_mask=_get_circle_by_center_bbox(shape,center,bbox,radius)
    return tmp_mask,bbox,center,radius


def mask_image(img,mask):
    img[mask<=0,...]=0
    return img


def remove_back_area(img,bbox=None,border=None):
    image=img
    if border is None:
        border=np.array((bbox[0],bbox[0]+bbox[2],bbox[1],bbox[1]+bbox[3],img.shape[0],img.shape[1]),dtype=np.int)
    image=image[border[0]:border[1],border[2]:border[3],...]
    return image,border


def supplemental_black_area(img,border=None):
    image=img
    if border is None:
        h,v=img.shape[0:2]
        max_l=max(h,v)
        if image.ndim>2:
            image=np.zeros(shape=[max_l,max_l,img.shape[2]],dtype=img.dtype)
        else:
            image=np.zeros(shape=[max_l,max_l],dtype=img.dtype)
        border=(int(max_l/2-h/2),int(max_l/2-h/2)+h,int(max_l/2-v/2),int(max_l/2-v/2)+v,max_l)
    else:
        max_l=border[4]
        if image.ndim>2:
            image=np.zeros(shape=[max_l,max_l,img.shape[2]],dtype=img.dtype)
        else:
            image=np.zeros(shape=[max_l,max_l],dtype=img.dtype)
    image[border[0]:border[1],border[2]:border[3],...]=img
    return image,border


def process_without_gb(img):
    # preprocess images
    #   img : origin image
    #   tar_height: height of tar image
    # return:
    #   result_img: preprocessed image
    #   borders: remove border, supplement mask
    #   mask: mask for preprocessed image
    borders = []
    mask, bbox, center, radius = get_mask(img)
    r_img = mask_image(img, mask)
    r_img, r_border = remove_back_area(r_img,bbox=bbox)
    mask, _ = remove_back_area(mask,border=r_border)
    borders.append(r_border)
    r_img,sup_border = supplemental_black_area(r_img)
    mask,_ = supplemental_black_area(mask,border=sup_border)
    borders.append(sup_border)
    return r_img,borders,(mask*255).astype(np.uint8)




import fundus_prep as prep
import glob
import os
import cv2 as cv
from PIL import ImageFile
ImageFile.LOAD_TRUNCATED_IMAGES = True

def process(image_list, save_path, extension):
    
    for image_path in image_list:
        dst_image = os.path.splitext(image_path.split('/')[-1])[0]+extension#'.png'
        dst_path = os.path.join(save_path, dst_image)
        if os.path.exists(dst_path):
            print('continue...')
            continue
        try:
            img = prep.imread(image_path)
            r_img, borders, mask = prep.process_without_gb(img)
            # r_img = cv.resize(r_img, (800, 800))
            prep.imwrite(dst_path, r_img)
            # mask = cv.resize(mask, (800, 800))
            # prep.imwrite(os.path.join('./original_mask', dst_image), mask)
        except:
            print(image_path)
            continue


if __name__ == "__main__":

    for i in range(3):
        image_list = glob.glob(os.path.join(f'/mnt/sda2/datasets/eyeq/test/{i}', '*.jpeg'))
        os.makedirs(f'/mnt/sda2/datasets/eyeq/test_preprocess/{i}', exist_ok = True)
        save_path = prep.fold_dir(f'/mnt/sda2/datasets/eyeq/test_preprocess/{i}')
        print(image_list[:10], save_path)
        process(image_list, save_path, ".jpg")

    #image_list = glob.glob(os.path.join('./original_img', '*.jpeg'))
    #save_path = prep.fold_dir('./original_crop')

    #process(image_list, save_path)

        





